"""Code generated from qulacs_worker namespace. Please do not edit."""

from typing import NamedTuple, Union
from types import NoneType
from tierkreis.controller.data.models import TKR, OpaqueType


class get_compiled_circuit(NamedTuple):
    circuit: TKR[OpaqueType["pytket._tket.circuit.Circuit"]]  # noqa: F821 # fmt: skip
    optimisation_level: TKR[int] | None = None  # noqa: F821 # fmt: skip
    result_type: TKR[str] | None = None  # noqa: F821 # fmt: skip
    gpu_sim: TKR[bool] | None = None  # noqa: F821 # fmt: skip

    @staticmethod
    def out() -> type[TKR[OpaqueType["pytket._tket.circuit.Circuit"]]]:  # noqa: F821 # fmt: skip
        return TKR[OpaqueType["pytket._tket.circuit.Circuit"]]  # noqa: F821 # fmt: skip

    @property
    def namespace(self) -> str:
        return "qulacs_worker"


class run_circuit(NamedTuple):
    circuit: TKR[OpaqueType["pytket._tket.circuit.Circuit"]]  # noqa: F821 # fmt: skip
    n_shots: TKR[int]  # noqa: F821 # fmt: skip
    result_type: TKR[str] | None = None  # noqa: F821 # fmt: skip
    gpu_sim: TKR[bool] | None = None  # noqa: F821 # fmt: skip
    seed: TKR[Union[int, NoneType]] | None = None  # noqa: F821 # fmt: skip

    @staticmethod
    def out() -> type[TKR[OpaqueType["pytket.backends.backendresult.BackendResult"]]]:  # noqa: F821 # fmt: skip
        return TKR[OpaqueType["pytket.backends.backendresult.BackendResult"]]  # noqa: F821 # fmt: skip

    @property
    def namespace(self) -> str:
        return "qulacs_worker"


class run_circuits(NamedTuple):
    circuits: TKR[list[OpaqueType["pytket._tket.circuit.Circuit"]]]  # noqa: F821 # fmt: skip
    n_shots: TKR[list[int]]  # noqa: F821 # fmt: skip
    result_type: TKR[str] | None = None  # noqa: F821 # fmt: skip
    gpu_sim: TKR[bool] | None = None  # noqa: F821 # fmt: skip
    seed: TKR[Union[int, NoneType]] | None = None  # noqa: F821 # fmt: skip

    @staticmethod
    def out() -> type[TKR[list[OpaqueType["pytket.backends.backendresult.BackendResult"]]]]:  # noqa: F821 # fmt: skip
        return TKR[list[OpaqueType["pytket.backends.backendresult.BackendResult"]]]  # noqa: F821 # fmt: skip

    @property
    def namespace(self) -> str:
        return "qulacs_worker"



from typing import NamedTuple
from tierkreis.builder import GraphBuilder
from tierkreis.controller.data.core import EmptyModel
from tierkreis.controller.data.graph import GraphData
from tierkreis.controller.data.models import TKR
from tierkreis.builtins.stubs import igt, iadd


def _loop_body_multiple_acc_untyped() -> GraphData:
    g = GraphData()

    acc = g.input("acc1")
    acc2 = g.input("acc2")
    acc3 = g.input("acc3")

    one = g.const(1)
    two = g.const(2)
    three = g.const(3)
    five = g.const(5)

    should_continue = g.func("builtins.igt", {"a": five, "b": acc})("value")

    new_acc = g.func("builtins.iadd", {"a": acc, "b": one})("value")
    new_acc2 = g.func("builtins.iadd", {"a": acc2, "b": two})("value")
    new_acc3 = g.func("builtins.iadd", {"a": acc3, "b": three})("value")

    g.output(
        {
            "should_continue": should_continue,
            "acc1": new_acc,
            "acc2": new_acc2,
            "acc3": new_acc3,
        }
    )

    return g


def loop_multiple_acc_untyped() -> GraphData:
    g = GraphData()

    acc1 = g.const(0)
    acc2 = g.const(0)
    acc3 = g.const(0)

    body_const = g.const(_loop_body_multiple_acc_untyped())

    loop = g.loop(
        body_const, {"acc1": acc1, "acc2": acc2, "acc3": acc3}, "should_continue"
    )

    g.output({"acc1": loop("acc1"), "acc2": loop("acc2"), "acc3": loop("acc3")})

    return g


class MultipleAcc(NamedTuple):
    acc1: TKR[int]
    acc2: TKR[int]
    acc3: TKR[int]


class MultipleAccOut(NamedTuple):
    should_continue: TKR[bool]
    acc1: TKR[int]
    acc2: TKR[int]
    acc3: TKR[int]


def _loop_body_multiple_acc() -> GraphBuilder[MultipleAcc, MultipleAccOut]:
    g = GraphBuilder(MultipleAcc, MultipleAccOut)

    acc = g.inputs.acc1
    acc2 = g.inputs.acc2
    acc3 = g.inputs.acc3

    one = g.const(1)
    two = g.const(2)
    three = g.const(3)
    five = g.const(5)

    should_continue = g.task(igt(a=five, b=acc))

    new_acc = g.task(iadd(a=acc, b=one))
    new_acc2 = g.task(iadd(a=acc2, b=two))
    new_acc3 = g.task(iadd(a=acc3, b=three))

    g.outputs(
        MultipleAccOut(
            should_continue=should_continue,
            acc1=new_acc,
            acc2=new_acc2,
            acc3=new_acc3,
        ),
    )

    return g


class LoopMultipleAccOut(NamedTuple):
    acc1: TKR[int]
    acc2: TKR[int]
    acc3: TKR[int]


def loop_multiple_acc() -> GraphBuilder[EmptyModel, LoopMultipleAccOut]:
    g = GraphBuilder(EmptyModel, LoopMultipleAccOut)

    acc1 = g.const(0)
    acc2 = g.const(0)
    acc3 = g.const(0)

    body = _loop_body_multiple_acc()
    loop = g.loop(body, MultipleAcc(acc1, acc2, acc3))

    g.outputs(LoopMultipleAccOut(acc1=loop.acc1, acc2=loop.acc2, acc3=loop.acc3))

    return g



from dataclasses import dataclass
from types import NoneType
from typing import Annotated, Mapping, Self, Sequence, TypeVar, get_args, get_origin

from tierkreis.controller.data.types import _is_generic
from tierkreis.exceptions import TierkreisError

type ElementaryType = (
    type[int]
    | type[float]
    | type[bytes]
    | type[str]
    | type[bool]
    | type[NoneType]
    | type[Mapping]
    | type[Sequence]
    | str  # Custom type e.g. forward reference
)


@dataclass
class GenericType:
    origin: ElementaryType
    args: "Sequence[GenericType | str]"

    @classmethod
    def from_type(cls, t: type) -> "Self":
        if get_origin(t) is Annotated:
            return cls.from_type(get_args(t)[0])

        if _is_generic(t):
            return cls(str(t), [])

        args, origin = get_args(t), get_origin(t)
        if not args:
            return cls(t, [])

        if args and origin:
            subargs = []
            for arg in args:
                if isinstance(arg, TypeVar) or isinstance(arg, TypeVar):
                    subargs.append(str(arg))
                else:
                    subargs.append(cls.from_type(arg))

            return cls(origin, subargs)

        raise TierkreisError(f"Expected generic type. Got {t}")

    @staticmethod
    def _generics(t: "GenericType | str"):
        if _is_generic(t):
            return str(t)

        if isinstance(t, str):
            return [t]

        if t.args == []:
            return []

        outs = []
        for arg in t.args:
            outs.extend(GenericType._generics(arg))
        return outs

    def generics(self) -> list[str]:
        return GenericType._generics(self)


@dataclass
class TypedArg:
    name: str
    t: GenericType
    has_default: bool = False


@dataclass
class Method:
    name: GenericType
    args: list[TypedArg]
    return_type: GenericType
    return_type_is_portmapping: bool = False


@dataclass
class Interface:
    name: str
    methods: list[Method]


@dataclass
class Model:
    is_portmapping: bool
    t: GenericType
    decls: list[TypedArg]

    def __hash__(self) -> int:
        return hash(self.t.origin)



from inspect import Signature, signature
import logging
from logging import getLogger
from pathlib import Path
import sys
from types import TracebackType
from typing import Callable

from tierkreis.controller.data.core import PortID
from tierkreis.controller.data.location import WorkerCallArgs
from tierkreis.controller.data.models import PModel, dict_from_pmodel
from tierkreis.controller.data.types import (
    PType,
    bytes_from_ptype,
    has_default,
    ptype_from_bytes,
)
from tierkreis.exceptions import TierkreisError
from tierkreis.namespace import Namespace, WorkerFunction
from tierkreis.worker.storage.filestorage import WorkerFileStorage
from tierkreis.worker.storage.protocol import WorkerStorage

logger = getLogger(__name__)
PrimitiveTask = Callable[[WorkerCallArgs, WorkerStorage], None]
type MethodName = str


def handle_unhandled_exception(
    exc_type: type[BaseException],
    exc_value: BaseException,
    exc_traceback: TracebackType | None,
):
    logger.critical(
        "Unhandled exception", exc_info=(exc_type, exc_value, exc_traceback)
    )


class TierkreisWorkerError(TierkreisError):
    pass


class Worker:
    """A worker bundles a set of functionality under a common namespace.

    The main usage of a worker is to convert python functions into atomic tasks,
    which can then be executed by the :py:class:`tierkreis.controller.executor.uv_executor.UvExecutor`
    or similar Executors.
    From the worker type stubs can be generated to statically check the function calls.

    Example snippet for wrapping a python function:

    .. code-block:: python

        worker = Worker("numpy_worker")

        @worker.function()
        def exp(x: float, a: float) -> float:
            return value = a * np.exp(x)

    :param name: The name of the worker.
    :type name: str
    :param storage: Storage layer for the worker to interact with the ControllerStorage.
    :type storage: WorkerStorage
    """

    functions: dict[str, Callable[[WorkerCallArgs], None]]
    types: dict[MethodName, Signature]
    namespace: Namespace

    def __init__(self, name: str, storage: WorkerStorage | None = None) -> None:
        self.name = name
        self.functions = {}
        self.types = {}
        self.namespace = Namespace(name=self.name, methods=[])
        if storage is None:
            self.storage: WorkerStorage = WorkerFileStorage()
        else:
            self.storage = storage
        sys.excepthook = handle_unhandled_exception

    def _load_args(
        self, f: WorkerFunction, inputs: dict[str, Path]
    ) -> dict[str, PType]:
        bs: dict[str, bytes] = {}
        for k, p in inputs.items():
            try:
                bs[k] = self.storage.read_input(p)
            except FileNotFoundError:
                if not has_default(self.types[f.__name__].parameters[k]):
                    raise TierkreisError(f"Input {k} not found at {p}.")

        args = {}
        for k, b in bs.items():
            args[k] = ptype_from_bytes(
                b, self.types[f.__name__].parameters[k].annotation
            )
        return args

    def _save_results(self, outputs: dict[PortID, Path], results: PModel):
        d = dict_from_pmodel(results)
        for result_name, path in outputs.items():
            self.storage.write_output(path, bytes_from_ptype(d[result_name]))

    def add_types(self, func: WorkerFunction) -> None:
        self.types[func.__name__] = signature(func)

    def primitive_task(
        self,
    ) -> Callable[[PrimitiveTask], None]:
        """Registers a python function as a primitive task with the worker."""

        def function_decorator(func: PrimitiveTask) -> None:
            def wrapper(args: WorkerCallArgs):
                func(args, self.storage)

            self.functions[func.__name__] = wrapper

        return function_decorator

    def task(self, name: str | None = None) -> Callable[[WorkerFunction], None]:
        """Registers a python function as a task with the worker."""

        def function_decorator(func: WorkerFunction) -> None:
            self.namespace.add_function(func)
            self.add_types(func)

            def wrapper(node_definition: WorkerCallArgs):
                kwargs = self._load_args(func, node_definition.inputs)
                results = func(**kwargs)
                self._save_results(node_definition.outputs, results)

            self.functions[func.__name__] = wrapper

        return function_decorator

    def run(self, worker_definition_path: Path) -> None:
        """Run a function with the parameters defined in worker_definition_path.

        :param worker_definition_path: The worker call args written by the controller.
        :type worker_definition_path: Path
        :raises TierkreisError: When the function execution results in an error.
        """
        node_definition = self.storage.read_call_args(worker_definition_path)

        logs_path = node_definition.logs_path
        logging.basicConfig(
            format="%(asctime)s: %(message)s",
            datefmt="%Y-%m-%dT%H:%M:%S%z",
            filename=self.storage.resolve(logs_path) if logs_path else None,
            filemode="a",
            level=logging.INFO,
        )
        logger.info(node_definition.model_dump())

        try:
            function = self.functions.get(node_definition.function_name, None)
            if function is None:
                raise TierkreisError(
                    f"{self.name}: function name {node_definition.function_name} not found"
                )
            logger.info(f"running: {node_definition.function_name} in {self.name}")

            function(node_definition)

            self.storage.mark_done(node_definition.done_path)
        except Exception as err:
            logger.error("encountered error", exc_info=err)
            self.storage.write_error(node_definition.error_path, str(err))

    def app(self, argv: list[str]) -> None:
        """Wrapper for UV execution."""
        if argv[1] == "--stubs-path":
            self.namespace.write_stubs(Path(argv[2]))
        else:
            self.run(Path(argv[1]))



from pathlib import Path
import uuid
import logging

from tierkreis.controller import run_graph
from tierkreis.controller.data.graph import GraphData
from tierkreis.controller.data.location import Loc
from tierkreis.controller.data.types import PType
from tierkreis.controller.storage.filestorage import ControllerFileStorage
from tierkreis.controller.executor.shell_executor import ShellExecutor
from tierkreis.controller.executor.uv_executor import UvExecutor

logger = logging.getLogger(__name__)


def run_workflow(
    graph: GraphData,
    inputs: dict[str, PType],
    name: str | None = None,
    run_id: int | None = None,
    log_level: int | str = logging.INFO,
    registry_path: Path | None = None,
    print_output: bool = False,
    use_uv_worker: bool = False,
    n_iterations: int = 10000,
    polling_interval_seconds: float = 0.1,
) -> None:
    """Run a workflow."""
    logging.basicConfig(
        format="%(asctime)s: %(message)s",
        datefmt="%Y-%m-%dT%H:%M:%S%z",
        level=log_level,
    )

    if run_id is None:
        workflow_id = uuid.uuid4()
    else:
        workflow_id = uuid.UUID(int=run_id)
    logging.info("Workflow ID is %s", workflow_id)
    storage = ControllerFileStorage(workflow_id, name=name, do_cleanup=True)
    if registry_path is None:
        registry_path = Path(__file__).parent
    if use_uv_worker:
        executor = UvExecutor(registry_path=registry_path, logs_path=storage.logs_path)
    else:
        executor = ShellExecutor(registry_path, storage.workflow_dir)

    logging.info("Starting workflow at location: %s", storage.logs_path)

    run_graph(
        storage,
        executor,
        graph,
        inputs,
        n_iterations,
        polling_interval_seconds,
    )
    if print_output:
        all_outputs = graph.nodes[graph.output_idx()].inputs
        for output in all_outputs:
            print(f"{output}: {storage.read_output(Loc(), output)}")



from dataclasses import dataclass
from inspect import isclass
from typing import Any, Callable, NamedTuple, Protocol, overload, runtime_checkable

from tierkreis.controller.data.core import EmptyModel
from tierkreis.controller.data.models import (
    TKR,
    TModel,
    TNamedModel,
    dict_from_tmodel,
    model_fields,
    init_tmodel,
)
from tierkreis.controller.data.types import PType
from tierkreis.controller.data.graph import GraphData, ValueRef


@dataclass
class TList[T: TModel]:
    """A list of models."""

    _value: T


@runtime_checkable
class Function[Out](TNamedModel, Protocol):
    @property
    def namespace(self) -> str: ...

    @staticmethod
    def out() -> type[Out]: ...


@dataclass
class TypedGraphRef[Ins: TModel, Outs: TModel]:
    graph_ref: ValueRef
    outputs_type: type[Outs]
    inputs_type: type[Ins]


class LoopOutput(TNamedModel, Protocol):
    @property
    def should_continue(self) -> TKR[bool]: ...


def script(script_name: str, input: TKR[bytes]) -> Function[TKR[bytes]]:
    class exec_script(NamedTuple):
        input: TKR[bytes]

        @staticmethod
        def out() -> type[TKR[bytes]]:
            return TKR[bytes]

        @property
        def namespace(self) -> str:
            return script_name

    return exec_script(input=input)


class GraphBuilder[Inputs: TModel, Outputs: TModel]:
    outputs_type: type
    inputs: Inputs

    def __init__(
        self,
        inputs_type: type[Inputs] = EmptyModel,
        outputs_type: type[Outputs] = EmptyModel,
    ):
        self.data = GraphData()
        self.inputs_type = inputs_type
        self.outputs_type = outputs_type
        inputs = [self.data.input(x) for x in model_fields(inputs_type)]
        self.inputs = init_tmodel(self.inputs_type, inputs)

    def get_data(self) -> GraphData:
        return self.data

    def ref(self) -> TypedGraphRef[Inputs, Outputs]:
        return TypedGraphRef((-1, "body"), self.outputs_type, self.inputs_type)

    def outputs(self, outputs: Outputs):
        self.data.output(inputs=dict_from_tmodel(outputs))

    def const[T: PType](self, value: T) -> TKR[T]:
        idx, port = self.data.const(value)
        return TKR[T](idx, port)

    def ifelse[A: PType, B: PType](
        self, pred: TKR[bool], if_true: TKR[A], if_false: TKR[B]
    ) -> TKR[A] | TKR[B]:
        idx, port = self.data.if_else(
            pred.value_ref(), if_true.value_ref(), if_false.value_ref()
        )("value")
        return TKR(idx, port)

    def eifelse[A: PType, B: PType](
        self, pred: TKR[bool], if_true: TKR[A], if_false: TKR[B]
    ) -> TKR[A] | TKR[B]:
        idx, port = self.data.eager_if_else(
            pred.value_ref(), if_true.value_ref(), if_false.value_ref()
        )("value")
        return TKR(idx, port)

    def _graph_const[A: TModel, B: TModel](
        self, graph: "GraphBuilder[A, B]"
    ) -> TypedGraphRef[A, B]:
        idx, port = self.data.const(graph.data.model_dump())
        return TypedGraphRef[A, B](
            graph_ref=(idx, port),
            outputs_type=graph.outputs_type,
            inputs_type=graph.inputs_type,
        )

    def task[Out: TModel](self, f: Function[Out]) -> Out:
        name = f"{f.namespace}.{f.__class__.__name__}"
        ins = dict_from_tmodel(f)
        idx, _ = self.data.func(name, ins)("dummy")
        OutModel = f.out()
        outputs = [(idx, x) for x in model_fields(OutModel)]
        return init_tmodel(OutModel, outputs)

    @overload
    def eval[A: TModel, B: TModel](self, body: TypedGraphRef[A, B], a: A) -> B: ...
    @overload
    def eval[A: TModel, B: TModel](self, body: "GraphBuilder[A, B]", a: A) -> B: ...
    def eval[A: TModel, B: TModel](
        self, body: "GraphBuilder[A,B] | TypedGraphRef", a: Any
    ) -> Any:
        if isinstance(body, GraphBuilder):
            body = self._graph_const(body)

        idx, _ = self.data.eval(body.graph_ref, dict_from_tmodel(a))("dummy")
        outputs = [(idx, x) for x in model_fields(body.outputs_type)]
        return init_tmodel(body.outputs_type, outputs)

    @overload
    def loop[A: TModel, B: LoopOutput](self, body: TypedGraphRef[A, B], a: A) -> B: ...
    @overload
    def loop[A: TModel, B: LoopOutput](self, body: "GraphBuilder[A, B]", a: A) -> B: ...
    def loop[A: TModel, B: LoopOutput](
        self, body: "TypedGraphRef[A, B] |GraphBuilder[A, B]", a: A
    ) -> B:
        if isinstance(body, GraphBuilder):
            body = self._graph_const(body)

        g = body.graph_ref
        idx, _ = self.data.loop(g, dict_from_tmodel(a), "should_continue")("dummy")
        outputs = [(idx, x) for x in model_fields(body.outputs_type)]
        return init_tmodel(body.outputs_type, outputs)

    def _unfold_list[T: PType](self, ref: TKR[list[T]]) -> TList[TKR[T]]:
        ins = (ref.node_index, ref.port_id)
        idx, _ = self.data.func("builtins.unfold_values", {"value": ins})("dummy")
        return TList(TKR[T](idx, "*"))

    def _fold_list[T: PType](self, refs: TList[TKR[T]]) -> TKR[list[T]]:
        value_ref = (refs._value.node_index, refs._value.port_id)
        idx, _ = self.data.func("builtins.fold_values", {"values_glob": value_ref})(
            "dummy"
        )
        return TKR[list[T]](idx, "value")

    def _map_fn_single_in[A: PType, B: TModel](
        self, aes: TKR[list[A]], body: Callable[[TKR[A]], B]
    ) -> "TList[B]":
        tlist = self._unfold_list(aes)
        return TList(body(TKR(tlist._value.node_index, "*")))

    def _map_fn_single_out[A: TModel, B: PType](
        self, aes: TList[A], body: Callable[[A], TKR[B]]
    ) -> TKR[list[B]]:
        return self._fold_list(TList(body(aes._value)))

    def _map_graph_full[A: TModel, B: TModel](
        self, aes: TList[A], body: TypedGraphRef[A, B]
    ) -> TList[B]:
        ins = dict_from_tmodel(aes._value)
        idx, _ = self.data.map(body.graph_ref, ins)("x")

        refs = [(idx, s + "-*") for s in model_fields(body.outputs_type)]
        return TList(init_tmodel(body.outputs_type, refs))

    @overload
    def map[A: PType, B: TNamedModel](
        self,
        body: (
            Callable[[TKR[A]], B] | TypedGraphRef[TKR[A], B] | "GraphBuilder[TKR[A], B]"
        ),
        aes: TKR[list[A]],
    ) -> TList[B]: ...

    @overload
    def map[A: TNamedModel, B: PType](
        self,
        body: (
            Callable[[A], TKR[B]] | TypedGraphRef[A, TKR[B]] | "GraphBuilder[A, TKR[B]]"
        ),
        aes: TList[A],
    ) -> TKR[list[B]]: ...

    @overload
    def map[A: TNamedModel, B: TNamedModel](
        self, body: TypedGraphRef[A, B] | "GraphBuilder[A, B]", aes: TList[A]
    ) -> TList[B]: ...

    @overload
    def map[A: PType, B: PType](
        self,
        body: (
            Callable[[TKR[A]], TKR[B]]
            | TypedGraphRef[TKR[A], TKR[B]]
            | "GraphBuilder[TKR[A], TKR[B]]"
        ),
        aes: TKR[list[A]],
    ) -> TKR[list[B]]: ...

    def map(
        self, body: TypedGraphRef | Callable | "GraphBuilder", aes: TKR | TList
    ) -> Any:
        if isinstance(body, GraphBuilder):
            body = self._graph_const(body)

        if isinstance(body, Callable):
            if isinstance(aes, TList):
                return self._map_fn_single_out(aes, body)
            elif isinstance(aes, TKR):
                return self._map_fn_single_in(aes, body)

        if isinstance(aes, TKR):
            aes = self._unfold_list(aes)

        out = self._map_graph_full(aes, body)

        if not isclass(body.outputs_type) or not issubclass(
            body.outputs_type, TNamedModel
        ):
            out = self._fold_list(out)

        return out



import pytest
import json
import logging
from pathlib import Path
from uuid import UUID
from unittest import mock


from tierkreis.controller.data.graph import GraphData
from tierkreis.cli.run_workflow import run_workflow
from tests.controller.sample_graphdata import simple_eval
from tierkreis.controller.data.types import ptype_from_bytes

logger = logging.getLogger(__name__)


@pytest.fixture()
def graph() -> GraphData:
    return simple_eval()


def test_run_workflow(graph: GraphData) -> None:
    inputs = {}
    run_workflow(inputs=inputs, graph=graph, run_id=31415)
    with open(
        Path.home()
        / ".tierkreis"
        / "checkpoints"
        / str(UUID(int=31415))
        / "-/outputs/simple_eval_output",
        "rb",
    ) as fh:
        c = ptype_from_bytes(fh.read())

    assert c == 12


def test_run_workflow_with_output(graph: GraphData, capfd) -> None:
    inputs = {}
    run_workflow(inputs=inputs, graph=graph, run_id=31415, print_output=True)
    out, _ = capfd.readouterr()
    assert "simple_eval_output: b'12'\n" in out


@mock.patch("uuid.uuid4", return_value=UUID(int=31415))
def test_run_workflow_default_run_id(_, graph: GraphData) -> None:
    inputs = {}
    run_workflow(inputs=inputs, graph=graph)
    with open(
        Path.home()
        / ".tierkreis"
        / "checkpoints"
        / str(UUID(int=31415))
        / "-/outputs/simple_eval_output",
        "rb",
    ) as fh:
        c = json.loads(fh.read())
    assert c == 12


def test_run_workflow_uv_executor(graph: GraphData) -> None:
    inputs = {}
    run_workflow(
        inputs=inputs,
        graph=graph,
        run_id=31415,
        use_uv_worker=True,
        registry_path=Path("."),
    )



"""Code generated from builtins namespace. Please do not edit."""

from typing import NamedTuple
from tierkreis.controller.data.models import TKR
from tierkreis.controller.data.types import PType


class Headed[T: PType](NamedTuple):
    head: TKR[T]  # noqa: F821 # fmt: skip
    rest: TKR[list[T]]  # noqa: F821 # fmt: skip


class Untupled[U: PType, V: PType](NamedTuple):
    a: TKR[U]  # noqa: F821 # fmt: skip
    b: TKR[V]  # noqa: F821 # fmt: skip


class Unzipped[U: PType, V: PType](NamedTuple):
    a: TKR[list[U]]  # noqa: F821 # fmt: skip
    b: TKR[list[V]]  # noqa: F821 # fmt: skip


class iadd(NamedTuple):
    a: TKR[int]  # noqa: F821 # fmt: skip
    b: TKR[int]  # noqa: F821 # fmt: skip

    @staticmethod
    def out() -> type[TKR[int]]:  # noqa: F821 # fmt: skip
        return TKR[int]  # noqa: F821 # fmt: skip

    @property
    def namespace(self) -> str:
        return "builtins"


class itimes(NamedTuple):
    a: TKR[int]  # noqa: F821 # fmt: skip
    b: TKR[int]  # noqa: F821 # fmt: skip

    @staticmethod
    def out() -> type[TKR[int]]:  # noqa: F821 # fmt: skip
        return TKR[int]  # noqa: F821 # fmt: skip

    @property
    def namespace(self) -> str:
        return "builtins"


class igt(NamedTuple):
    a: TKR[int]  # noqa: F821 # fmt: skip
    b: TKR[int]  # noqa: F821 # fmt: skip

    @staticmethod
    def out() -> type[TKR[bool]]:  # noqa: F821 # fmt: skip
        return TKR[bool]  # noqa: F821 # fmt: skip

    @property
    def namespace(self) -> str:
        return "builtins"


class neg(NamedTuple):
    a: TKR[bool]  # noqa: F821 # fmt: skip

    @staticmethod
    def out() -> type[TKR[bool]]:  # noqa: F821 # fmt: skip
        return TKR[bool]  # noqa: F821 # fmt: skip

    @property
    def namespace(self) -> str:
        return "builtins"


class impl_and(NamedTuple):
    a: TKR[bool]  # noqa: F821 # fmt: skip
    b: TKR[bool]  # noqa: F821 # fmt: skip

    @staticmethod
    def out() -> type[TKR[bool]]:  # noqa: F821 # fmt: skip
        return TKR[bool]  # noqa: F821 # fmt: skip

    @property
    def namespace(self) -> str:
        return "builtins"


class impl_id[T: PType](NamedTuple):
    value: TKR[T]  # noqa: F821 # fmt: skip

    @staticmethod
    def out() -> type[TKR[T]]:  # noqa: F821 # fmt: skip
        return TKR[T]  # noqa: F821 # fmt: skip

    @property
    def namespace(self) -> str:
        return "builtins"


class append[T: PType](NamedTuple):
    v: TKR[list[T]]  # noqa: F821 # fmt: skip
    a: TKR[T]  # noqa: F821 # fmt: skip

    @staticmethod
    def out() -> type[TKR[list[T]]]:  # noqa: F821 # fmt: skip
        return TKR[list[T]]  # noqa: F821 # fmt: skip

    @property
    def namespace(self) -> str:
        return "builtins"


class head[T: PType](NamedTuple):
    v: TKR[list[T]]  # noqa: F821 # fmt: skip

    @staticmethod
    def out() -> type[Headed[T]]:  # noqa: F821 # fmt: skip
        return Headed[T]  # noqa: F821 # fmt: skip

    @property
    def namespace(self) -> str:
        return "builtins"


class impl_len[A: PType](NamedTuple):
    v: TKR[list[A]]  # noqa: F821 # fmt: skip

    @staticmethod
    def out() -> type[TKR[int]]:  # noqa: F821 # fmt: skip
        return TKR[int]  # noqa: F821 # fmt: skip

    @property
    def namespace(self) -> str:
        return "builtins"


class str_eq(NamedTuple):
    a: TKR[str]  # noqa: F821 # fmt: skip
    b: TKR[str]  # noqa: F821 # fmt: skip

    @staticmethod
    def out() -> type[TKR[bool]]:  # noqa: F821 # fmt: skip
        return TKR[bool]  # noqa: F821 # fmt: skip

    @property
    def namespace(self) -> str:
        return "builtins"


class str_neq(NamedTuple):
    a: TKR[str]  # noqa: F821 # fmt: skip
    b: TKR[str]  # noqa: F821 # fmt: skip

    @staticmethod
    def out() -> type[TKR[bool]]:  # noqa: F821 # fmt: skip
        return TKR[bool]  # noqa: F821 # fmt: skip

    @property
    def namespace(self) -> str:
        return "builtins"


class concat(NamedTuple):
    lhs: TKR[str]  # noqa: F821 # fmt: skip
    rhs: TKR[str]  # noqa: F821 # fmt: skip

    @staticmethod
    def out() -> type[TKR[str]]:  # noqa: F821 # fmt: skip
        return TKR[str]  # noqa: F821 # fmt: skip

    @property
    def namespace(self) -> str:
        return "builtins"


class zip_impl[U: PType, V: PType](NamedTuple):
    a: TKR[list[U]]  # noqa: F821 # fmt: skip
    b: TKR[list[V]]  # noqa: F821 # fmt: skip

    @staticmethod
    def out() -> type[TKR[list[tuple[U, V]]]]:  # noqa: F821 # fmt: skip
        return TKR[list[tuple[U, V]]]  # noqa: F821 # fmt: skip

    @property
    def namespace(self) -> str:
        return "builtins"


class unzip[U: PType, V: PType](NamedTuple):
    value: TKR[list[tuple[U, V]]]  # noqa: F821 # fmt: skip

    @staticmethod
    def out() -> type[Unzipped[U, V]]:  # noqa: F821 # fmt: skip
        return Unzipped[U, V]  # noqa: F821 # fmt: skip

    @property
    def namespace(self) -> str:
        return "builtins"


class tuple_impl[U: PType, V: PType](NamedTuple):
    a: TKR[U]  # noqa: F821 # fmt: skip
    b: TKR[V]  # noqa: F821 # fmt: skip

    @staticmethod
    def out() -> type[TKR[tuple[U, V]]]:  # noqa: F821 # fmt: skip
        return TKR[tuple[U, V]]  # noqa: F821 # fmt: skip

    @property
    def namespace(self) -> str:
        return "builtins"


class untuple[U: PType, V: PType](NamedTuple):
    value: TKR[tuple[U, V]]  # noqa: F821 # fmt: skip

    @staticmethod
    def out() -> type[Untupled[U, V]]:  # noqa: F821 # fmt: skip
        return Untupled[U, V]  # noqa: F821 # fmt: skip

    @property
    def namespace(self) -> str:
        return "builtins"


class mean(NamedTuple):
    values: TKR[list[float]]  # noqa: F821 # fmt: skip

    @staticmethod
    def out() -> type[TKR[float]]:  # noqa: F821 # fmt: skip
        return TKR[float]  # noqa: F821 # fmt: skip

    @property
    def namespace(self) -> str:
        return "builtins"


class mod(NamedTuple):
    a: TKR[int]  # noqa: F821 # fmt: skip
    b: TKR[int]  # noqa: F821 # fmt: skip

    @staticmethod
    def out() -> type[TKR[int]]:  # noqa: F821 # fmt: skip
        return TKR[int]  # noqa: F821 # fmt: skip

    @property
    def namespace(self) -> str:
        return "builtins"


class rand_int(NamedTuple):
    a: TKR[int]  # noqa: F821 # fmt: skip
    b: TKR[int]  # noqa: F821 # fmt: skip

    @staticmethod
    def out() -> type[TKR[int]]:  # noqa: F821 # fmt: skip
        return TKR[int]  # noqa: F821 # fmt: skip

    @property
    def namespace(self) -> str:
        return "builtins"


class tkr_sleep(NamedTuple):
    delay_seconds: TKR[float]  # noqa: F821 # fmt: skip

    @staticmethod
    def out() -> type[TKR[bool]]:  # noqa: F821 # fmt: skip
        return TKR[bool]  # noqa: F821 # fmt: skip

    @property
    def namespace(self) -> str:
        return "builtins"


class tkr_encode(NamedTuple):
    string: TKR[str]  # noqa: F821 # fmt: skip

    @staticmethod
    def out() -> type[TKR[bytes]]:  # noqa: F821 # fmt: skip
        return TKR[bytes]  # noqa: F821 # fmt: skip

    @property
    def namespace(self) -> str:
        return "builtins"


class tkr_decode(NamedTuple):
    bytes: TKR[bytes]  # noqa: F821 # fmt: skip

    @staticmethod
    def out() -> type[TKR[str]]:  # noqa: F821 # fmt: skip
        return TKR[str]  # noqa: F821 # fmt: skip

    @property
    def namespace(self) -> str:
        return "builtins"



import logging
from pathlib import Path

from tierkreis.controller.data.location import WorkerCallArgs
from tierkreis.controller.storage.in_memory import ControllerInMemoryStorage
from tierkreis.exceptions import TierkreisError


logger = logging.getLogger(__name__)


class InMemoryWorkerStorage:
    def __init__(self, controller_storage: ControllerInMemoryStorage) -> None:
        self.controller_storage = controller_storage

    def resolve(self, path: Path | str) -> Path:
        return Path(path)

    def read_call_args(self, path: Path) -> WorkerCallArgs:
        loc = self.controller_storage.path_to_loc(path)[0]
        return self.controller_storage.read_worker_call_args(loc)

    def read_input(self, path: Path) -> bytes:
        loc, port = self.controller_storage.path_to_loc(path)
        if port is None:
            raise TierkreisError(f"Path {path} does not specify a port.")
        return self.controller_storage.read_output(loc, port)

    def write_output(self, path: Path, value: bytes) -> None:
        loc, port = self.controller_storage.path_to_loc(path)
        if port is None:
            raise TierkreisError(f"Path {path} does not specify a port.")
        self.controller_storage.write_output(loc, port, value)

    def glob(self, path_string: str) -> list[str]:
        loc, port = self.controller_storage.path_to_loc(Path(path_string))
        path = self.controller_storage.loc_to_path(loc)
        outputs = [str(key) for key in self.controller_storage.nodes[path].outputs]
        if port is not None:
            outputs = filter(lambda o: o.startswith(port[:-1]), outputs)
        return [loc + "/" + out for out in outputs]

    def mark_done(self, path: Path) -> None:
        loc, _ = self.controller_storage.path_to_loc(path)
        self.controller_storage.mark_node_finished(loc)

    def write_error(self, path: Path, error_logs: str) -> None:
        loc, _ = self.controller_storage.path_to_loc(path)
        self.controller_storage.write_node_errors(loc, error_logs)



import json
import subprocess
from pathlib import Path

from tierkreis.consts import TKR_DIR_KEY
from tierkreis.controller.data.location import WorkerCallArgs
from tierkreis.exceptions import TierkreisError


class ShellExecutor:
    """Executes workers in an unix shell.

    Implements: :py:class:`tierkreis.controller.executor.protocol.ControllerExecutor`
    """

    def __init__(
        self,
        registry_path: Path,
        workflow_dir: Path,
        timeout: int = 10,
        env: dict[str, str] | None = None,
    ) -> None:
        self.launchers_path = registry_path
        self.logs_path = workflow_dir / "logs"
        self.errors_path = workflow_dir / "logs"
        self.workflow_dir = workflow_dir
        self.timeout = timeout
        self.env = env or {}

    def run(
        self,
        launcher_name: str,
        worker_call_args_path: Path,
        export_values: bool = False,
    ) -> None:
        launcher_path = self.launchers_path / launcher_name
        self.errors_path = worker_call_args_path.parent / "errors"

        if not launcher_path.exists():
            raise TierkreisError(f"Launcher not found: {launcher_name}.")

        if launcher_path.is_dir() and not (launcher_path / "main.sh").exists():
            raise TierkreisError(f"Expected launcher file. Got {launcher_path}.")

        if launcher_path.is_dir() and not (launcher_path / "main.sh").is_file():
            raise TierkreisError(f"Expected launcher file. Got {launcher_path}/main.sh")

        if launcher_path.is_dir() and (launcher_path / "main.sh").is_file():
            launcher_path = launcher_path / "main.sh"

        with open(self.workflow_dir.parent / worker_call_args_path) as fh:
            call_args = WorkerCallArgs(**json.load(fh))

        env = self.env.copy()
        env.update(self._create_env(call_args, self.workflow_dir.parent, export_values))
        env["worker_call_args_file"] = str(
            self.workflow_dir.parent / worker_call_args_path
        )
        done_path = self.workflow_dir.parent / call_args.done_path
        _error_path = self.errors_path.parent / "_error"
        if TKR_DIR_KEY not in env:
            env[TKR_DIR_KEY] = str(self.logs_path.parent.parent)

        with open(self.workflow_dir.parent / self.logs_path, "a") as lfh:
            with open(self.workflow_dir.parent / self.errors_path, "a") as efh:
                proc = subprocess.Popen(
                    ["bash"],
                    start_new_session=True,
                    stdin=subprocess.PIPE,
                    stderr=efh,
                    stdout=lfh,
                    env=env,
                )
                proc.communicate(
                    f"({launcher_path} {worker_call_args_path} && touch {done_path}|| touch {_error_path})&".encode(),
                    timeout=self.timeout,
                )

    def _create_env(
        self, call_args: WorkerCallArgs, base_dir: Path, export_values: bool
    ) -> dict[str, str]:
        env = {
            "checkpoints_directory": str(base_dir),
            "function_name": str(base_dir / call_args.function_name),
            "done_path": str(base_dir / call_args.done_path),
            "error_path": str(base_dir / call_args.error_path),
            "output_dir": str(base_dir / call_args.output_dir),
        }
        if call_args.logs_path is not None:
            env["logs_path"] = str(base_dir / call_args.logs_path)
        else:
            env["logs_path"] = str(self.logs_path)
        env |= {
            f"output_{k}_file": str(base_dir / v) for k, v in call_args.outputs.items()
        }
        env |= {
            f"input_{k}_file": str(base_dir / v) for k, v in call_args.inputs.items()
        }
        if not export_values:
            return env
        values = {}
        for k, v in call_args.inputs.items():
            with open(v) as fh:
                values[f"input_{k}_value"] = fh.read()
        return env



# /// script
# requires-python = ">=3.12"
# dependencies = ["pydantic", "tierkreis"]
#
# [tool.uv.sources]
# tierkreis = { path = "../../../tierkreis", editable = true }
# ///
from pathlib import Path
from time import sleep
from sys import argv

from tierkreis import Worker, Value

worker = Worker("tests_worker")


@worker.task()
def sleep_and_return[T](*, output: T) -> Value[T]:
    sleep(10)
    return Value(value=output)


def main() -> None:
    node_definition_path = argv[1]
    worker.run(Path(node_definition_path))


if __name__ == "__main__":
    main()



from pathlib import Path
from uuid import UUID
from typing import Any


from tierkreis.controller.data.core import PortID
from tierkreis.controller.data.graph import (
    Eval,
    GraphData,
    NodeDef,
    graph_node_from_loc,
)
from tierkreis.controller.data.location import (
    Loc,
    OutputLoc,
    WorkerCallArgs,
    get_last_index,
)
from tierkreis.exceptions import TierkreisError
from tierkreis.controller.storage.in_memory import NodeData


class GraphDataStorage:
    def __init__(
        self,
        workflow_id: UUID,
        graph: GraphData,
        name: str | None = None,
    ) -> None:
        self.workflow_id = workflow_id
        self.name = name
        self.nodes: dict[Loc, NodeData] = {}
        self.graph = graph
        self.logs_path = Path("")

    def write_node_def(self, node_location: Loc, node: NodeDef) -> None:
        raise NotImplementedError("GraphDataStorage is read only storage.")

    def read_node_def(self, node_location: Loc) -> NodeDef:
        try:
            if node_location.pop_last()[0][0] in ["M", "L"]:
                return Eval((-1, "body"), {})
        except (TierkreisError, TypeError):
            return Eval((-1, "body"), {})
        node, _ = graph_node_from_loc(node_location, self.graph)
        return node

    def write_worker_call_args(
        self,
        node_location: Loc,
        function_name: str,
        inputs: dict[PortID, OutputLoc],
        output_list: list[PortID],
    ) -> Path:
        raise NotImplementedError("GraphDataStorage is read only storage.")

    def read_worker_call_args(self, node_location: Loc) -> WorkerCallArgs:
        raise TierkreisError(
            f"Node location {node_location} doesn't have a associate call args."
        )

    def read_errors(self, node_location: Loc) -> str:
        return ""

    def node_has_error(self, node_location: Loc) -> bool:
        return False

    def write_node_errors(self, node_location: Loc, error_logs: str) -> None:
        raise NotImplementedError("GraphDataStorage is read only storage.")

    def mark_node_finished(self, node_location: Loc) -> None:
        raise NotImplementedError("GraphDataStorage is read only storage.")

    def is_node_finished(self, node_location: Loc) -> bool:
        return False

    def link_outputs(
        self,
        new_location: Loc,
        new_port: PortID,
        old_location: Loc,
        old_port: PortID,
    ) -> None:
        raise NotImplementedError("GraphDataStorage is read only storage.")

    def write_output(
        self, node_location: Loc, output_name: PortID, value: bytes
    ) -> Path:
        raise NotImplementedError("GraphDataStorage is read only storage.")

    def read_output(self, node_location: Loc, output_name: PortID) -> bytes:
        node, graph = graph_node_from_loc(node_location, self.graph)
        if -1 == get_last_index(node_location) and output_name == "body":
            return graph.model_dump_json().encode()

        outputs = _build_node_outputs(node)
        if output_name in outputs:
            if output := outputs[output_name]:
                return output
            return b"null"
        raise TierkreisError(f"No output named {output_name} in node {node_location}")

    def read_output_ports(self, node_location: Loc) -> list[PortID]:
        node, _ = graph_node_from_loc(node_location, self.graph)
        outputs = _build_node_outputs(node)
        return list(filter(lambda k: k != "*", outputs.keys()))

    def is_node_started(self, node_location: Loc) -> bool:
        return False

    def read_metadata(self, node_location: Loc) -> dict[str, Any]:
        return self.nodes[node_location].metadata

    def write_metadata(self, node_location: Loc) -> None:
        raise NotImplementedError("GraphDataStorage is read only storage.")

    def read_started_time(self, node_location: Loc) -> str | None:
        return None

    def read_finished_time(self, node_location: Loc) -> str | None:
        return None


def _build_node_outputs(node: NodeDef) -> dict[PortID, None | bytes]:
    outputs: dict[PortID, None | bytes] = {val: None for val in node.outputs}
    if "*" in outputs:
        outputs["0"] = None
    return outputs



from dataclasses import dataclass, field
from logging import getLogger
from typing import assert_never

from tierkreis.controller.consts import BODY_PORT
from tierkreis.controller.data.core import NodeIndex
from tierkreis.controller.data.graph import (
    EagerIfElse,
    Eval,
    GraphData,
    Loop,
    Map,
    NodeDef,
)
from tierkreis.controller.data.location import Loc
from tierkreis.controller.data.types import ptype_from_bytes
from tierkreis.controller.start import NodeRunData
from tierkreis.controller.storage.adjacency import outputs_iter, unfinished_inputs
from tierkreis.controller.storage.protocol import ControllerStorage
from tierkreis.labels import Labels

logger = getLogger(__name__)


@dataclass
class WalkResult:
    inputs_ready: list[NodeRunData]
    started: list[Loc]
    errored: list[Loc] = field(default_factory=list[Loc])

    def extend(self, walk_result: "WalkResult") -> None:
        self.inputs_ready.extend(walk_result.inputs_ready)
        self.started.extend(walk_result.started)
        self.errored.extend(walk_result.errored)


def unfinished_results(
    result: WalkResult,
    storage: ControllerStorage,
    parent: Loc,
    node: NodeDef,
    graph: GraphData,
) -> int:
    unfinished = unfinished_inputs(storage, parent, node)
    [result.extend(walk_node(storage, parent, x[0], graph)) for x in unfinished]
    return len(unfinished)


def walk_node(
    storage: ControllerStorage, parent: Loc, idx: NodeIndex, graph: GraphData
) -> WalkResult:
    """Should only be called when a node has not finished."""
    loc = parent.N(idx)
    if storage.node_has_error(loc):
        logger.error(f"Node {loc} has encountered an error:")
        logger.error(f"\n\n{storage.read_errors(loc)}\n\n")
        return WalkResult([], [], [loc])

    node = graph.nodes[idx]
    node_run_data = NodeRunData(loc, node, list(node.outputs))

    result = WalkResult([], [])
    if unfinished_results(result, storage, parent, node, graph):
        return result

    if not storage.is_node_started(loc):
        return WalkResult([node_run_data], [])

    match node.type:
        case "eval":
            message = storage.read_output(parent.N(node.graph[0]), node.graph[1])
            g = ptype_from_bytes(message, GraphData)
            return walk_node(storage, loc, g.output_idx(), g)

        case "output":
            return WalkResult([node_run_data], [])

        case "const":
            return WalkResult([node_run_data], [])

        case "loop":
            return walk_loop(storage, parent, idx, node)

        case "map":
            return walk_map(storage, parent, idx, node)

        case "ifelse":
            pred = storage.read_output(parent.N(node.pred[0]), node.pred[1])
            next_node = node.if_true if pred == b"true" else node.if_false
            next_loc = parent.N(next_node[0])
            if storage.is_node_finished(next_loc):
                storage.link_outputs(loc, Labels.VALUE, next_loc, next_node[1])
                storage.mark_node_finished(loc)
                return WalkResult([], [])
            else:
                return walk_node(storage, parent, next_node[0], graph)

        case "eifelse":
            return walk_eifelse(storage, parent, idx, node)

        case "function":
            return WalkResult([], [loc])

        case "input":
            return WalkResult([], [])
        case _:
            assert_never(node)


def walk_loop(
    storage: ControllerStorage, parent: Loc, idx: NodeIndex, loop: Loop
) -> WalkResult:
    loc = parent.N(idx)
    if storage.is_node_finished(loc):
        return WalkResult([], [], [])

    i = 0
    while storage.is_node_started(loc.L(i + 1)):
        i += 1
    new_location = loc.L(i)

    message = storage.read_output(loc.N(-1), BODY_PORT)
    g = ptype_from_bytes(message, GraphData)
    loop_outputs = g.nodes[g.output_idx()].inputs

    if not storage.is_node_finished(new_location):
        return walk_node(storage, new_location, g.output_idx(), g)

    # Latest iteration is finished. Do we BREAK or CONTINUE?
    should_continue = ptype_from_bytes(
        storage.read_output(new_location, loop.continue_port), bool
    )
    if should_continue is False:
        for k in loop_outputs:
            storage.link_outputs(loc, k, new_location, k)
        storage.mark_node_finished(loc)
        return WalkResult([], [])

    ins = {k: (-1, k) for k in loop.inputs.keys()}
    ins.update(loop_outputs)
    node_run_data = NodeRunData(
        loc.L(i + 1),
        Eval((-1, BODY_PORT), ins, loop.outputs),
        list(loop_outputs.keys()),
    )
    return WalkResult([node_run_data], [])


def walk_map(
    storage: ControllerStorage, parent: Loc, idx: NodeIndex, map: Map
) -> WalkResult:
    loc = parent.N(idx)
    result = WalkResult([], [])
    if storage.is_node_finished(loc):
        return result

    first_ref = next(x for x in map.inputs.values() if x[1] == "*")
    map_eles = outputs_iter(storage, parent.N(first_ref[0]))
    unfinished = [i for i, _ in map_eles if not storage.is_node_finished(loc.M(i))]
    message = storage.read_output(loc.M(0).N(-1), BODY_PORT)
    g = ptype_from_bytes(message, GraphData)
    [result.extend(walk_node(storage, loc.M(p), g.output_idx(), g)) for p in unfinished]

    if len(unfinished) > 0:
        return result

    map_outputs = g.nodes[g.output_idx()].inputs
    for i, j in map_eles:
        for output in map_outputs.keys():
            storage.link_outputs(loc, f"{output}-{j}", loc.M(i), output)

    storage.mark_node_finished(loc)
    return result


def walk_eifelse(
    storage: ControllerStorage,
    parent: Loc,
    idx: NodeIndex,
    node: EagerIfElse,
) -> WalkResult:
    loc = parent.N(idx)
    pred = storage.read_output(parent.N(node.pred[0]), node.pred[1])
    next_node = node.if_true if pred == b"true" else node.if_false
    next_loc = parent.N(next_node[0])
    storage.link_outputs(loc, Labels.VALUE, next_loc, next_node[1])
    storage.mark_node_finished(loc)

    return WalkResult([], [])



"""Combinators for the Tierkreis worker IDL.

We use
https://typespec.io/docs/language-basics/models/
https://typespec.io/docs/language-basics/interfaces/
as well as an extra decorator @portmapping.
"""

from tierkreis.idl.models import Interface, Method, Model, TypedArg

from tierkreis.idl.parser import lit, seq
from tierkreis.idl.type_symbols import generic_t, ident, type_symbol


type_decl = ((ident << lit(":")) & type_symbol).map(lambda x: TypedArg(*x))
model = seq(
    lit("@portmapping").opt().map(lambda x: x is not None) << lit("model"),
    generic_t,
    lit("{") >> type_decl.rep(lit(";")) << lit("}"),
).map(lambda x: Model(*x))
method = seq(
    generic_t,
    lit("(") >> type_decl.rep(lit(",")) << lit(")") << lit(":"),
    type_symbol,
).map(lambda x: Method(*x))
interface = (
    (lit("interface") >> ident << lit("{")) & method.rep(lit(";")) << lit("}")
).map(lambda x: Interface(*x))
spec = model.rep() & interface



from collections import defaultdict
from datetime import datetime
from pathlib import Path
from uuid import UUID
from typing import Any
import shutil
import os
from time import time_ns

from pydantic import BaseModel, Field

from tierkreis.controller.data.core import PortID
from tierkreis.controller.data.graph import NodeDef
from tierkreis.controller.data.location import (
    Loc,
    OutputLoc,
    WorkerCallArgs,
)
from tierkreis.exceptions import TierkreisError


class NodeData(BaseModel):
    """Internal storage class to store all necessary node information."""

    definition: NodeDef | None = None
    call_args: WorkerCallArgs | None = None
    is_done: bool = False
    has_error: bool = False
    metadata: dict[str, Any] = Field(default_factory=dict)
    error_logs: str = ""
    outputs: dict[PortID, bytes | None] = Field(default_factory=dict)
    started: str | None = None
    finished: str | None = None


class ControllerInMemoryStorage:
    """In-memory storage for the controller.

    All information is kept in a local dictionary.
    Invalid read operations raise a TierkreisError.

    Implements: :py:class:`tierkreis.controller.storage.protocol.ControllerStorage`
    """

    def __init__(
        self,
        workflow_id: UUID,
        name: str | None = None,
    ) -> None:
        self.work_dir = Path.home() / ".tierkreis" / "tmp"
        self.clean_graph_files()
        self.work_dir.mkdir(parents=True, exist_ok=True)
        self.workflow_id = workflow_id
        self.logs_path = self.work_dir / "logs"
        self.logs_path.touch()
        self.name = name
        self.nodes: dict[Path, NodeData] = defaultdict(lambda: NodeData())

    def loc_to_path(self, location: Loc, port: PortID | None = None) -> Path:
        # Directly convert, not a real filesystem path
        if port is not None:
            return Path(location) / port
        return Path(location)

    def path_to_loc(self, path: Path) -> tuple[Loc, PortID | None]:
        # assumes the path is in the format 'location/port'
        parts = path.parts
        if len(parts) == 1:
            return Loc(parts[0]), None
        elif len(parts) == 2:
            return Loc(parts[0]), PortID(parts[1])
        else:
            raise ValueError(f"Invalid path format: {path}")

    def write_node_def(self, node_location: Loc, node: NodeDef) -> None:
        self.nodes[self.loc_to_path(node_location)].definition = node
        self.nodes[self.loc_to_path(node_location)].started = datetime.now().isoformat()

    def read_node_def(self, node_location: Loc) -> NodeDef:
        if result := self.nodes[self.loc_to_path(node_location)].definition:
            return result
        raise TierkreisError(f"Node definition of {node_location} not found.")

    def write_worker_call_args(
        self,
        node_location: Loc,
        function_name: str,
        inputs: dict[PortID, OutputLoc],
        output_list: list[PortID],
    ) -> Path:
        node_path = self.loc_to_path(node_location)
        call_args = WorkerCallArgs(
            function_name=function_name,
            inputs={
                k: self.loc_to_path(loc, port) for k, (loc, port) in inputs.items()
            },
            outputs={k: self.loc_to_path(node_location, k) for k in output_list},
            output_dir=node_path,
            done_path=node_path,
            error_path=node_path,
            logs_path=self.logs_path,
        )
        for port in output_list:
            # workaround
            self.nodes[node_path].outputs[port] = None
        self.nodes[node_path].call_args = call_args
        if (parent := node_location.parent()) is not None:
            self.nodes[self.loc_to_path(parent)].metadata = {}

        return node_path

    def read_worker_call_args(self, node_location: Loc) -> WorkerCallArgs:
        if result := self.nodes[self.loc_to_path(node_location)].call_args:
            return result
        raise TierkreisError(
            f"Node location {node_location} doesn't have a associate call args."
        )

    def read_errors(self, node_location: Loc) -> str:
        if errors := self.nodes[self.loc_to_path(node_location)].error_logs:
            return errors
        return ""

    def node_has_error(self, node_location: Loc) -> bool:
        return self.nodes[self.loc_to_path(node_location)].has_error

    def write_node_errors(self, node_location: Loc, error_logs: str) -> None:
        self.nodes[self.loc_to_path(node_location)].error_logs = error_logs

    def mark_node_finished(self, node_location: Loc) -> None:
        self.nodes[self.loc_to_path(node_location)].is_done = True
        self.nodes[
            self.loc_to_path(node_location)
        ].finished = datetime.now().isoformat()

    def is_node_finished(self, node_location: Loc) -> bool:
        return self.nodes[self.loc_to_path(node_location)].is_done

    def link_outputs(
        self,
        new_location: Loc,
        new_port: PortID,
        old_location: Loc,
        old_port: PortID,
    ) -> None:
        self.nodes[self.loc_to_path(new_location)].outputs[new_port] = self.nodes[
            self.loc_to_path(old_location)
        ].outputs[old_port]

    def write_output(
        self, node_location: Loc, output_name: PortID, value: bytes
    ) -> Path:
        self.nodes[self.loc_to_path(node_location)].outputs[output_name] = value
        return Path(node_location) / output_name

    def read_output(self, node_location: Loc, output_name: PortID) -> bytes:
        if output_name in self.nodes[self.loc_to_path(node_location)].outputs:
            if output := self.nodes[self.loc_to_path(node_location)].outputs[
                output_name
            ]:
                return output
            return b"null"
        raise TierkreisError(f"No output named {output_name} in node {node_location}")

    def read_output_ports(self, node_location: Loc) -> list[PortID]:
        return list(
            filter(
                lambda k: k != "*",
                self.nodes[self.loc_to_path(node_location)].outputs.keys(),
            )
        )

    def is_node_started(self, node_location: Loc) -> bool:
        return self.nodes[self.loc_to_path(node_location)].definition is not None

    def read_metadata(self, node_location: Loc) -> dict[str, Any]:
        return self.nodes[self.loc_to_path(node_location)].metadata

    def write_metadata(self, node_location: Loc) -> None:
        self.nodes[self.loc_to_path(node_location)].metadata = {
            "name": self.name,
            "start_time": datetime.now().isoformat(),
        }

    def read_started_time(self, node_location: Loc) -> str | None:
        return self.nodes[self.loc_to_path(node_location)].started

    def read_finished_time(self, node_location: Loc) -> str | None:
        return self.nodes[self.loc_to_path(node_location)].finished

    def clean_graph_files(self) -> None:
        uid = os.getuid()
        tmp_dir = Path(f"/tmp/{uid}/tierkreis/archive/{time_ns()}")
        tmp_dir.mkdir(parents=True)
        if self.work_dir.exists():
            shutil.move(self.work_dir, tmp_dir)



"""Code generated from slurm_mpi_worker namespace. Please do not edit."""

from typing import NamedTuple
from types import NoneType
from tierkreis.controller.data.models import TKR


class mpi_rank_info(NamedTuple):
    @staticmethod
    def out() -> type[TKR[str | NoneType]]:  # noqa: F821 # fmt: skip
        return TKR[str | NoneType]  # noqa: F821 # fmt: skip

    @property
    def namespace(self) -> str:
        return "slurm_mpi_worker"



from pathlib import Path
import pytest
from tierkreis.exceptions import TierkreisError
from tierkreis.idl.models import GenericType
from tierkreis.namespace import Namespace
from tierkreis.idl.type_symbols import type_symbol
import tests.idl.namespace1

type_symbols = [
    ("uint8", GenericType(int, [])),
    ("string", GenericType(str, [])),
    ("Array<integer>", GenericType(list, [GenericType(int, [])])),
    (
        "Record<Array<string>>",
        GenericType(
            dict, [GenericType(str, []), GenericType(list, [GenericType(str, [])])]
        ),
    ),
    (
        "Record<Array<C<T, A>>>",
        GenericType(
            dict,
            [
                GenericType(str, []),
                GenericType(list, [GenericType("C", ["T", "A"])]),
            ],
        ),
    ),
]
type_symbols_for_failure = ["decimal", "unknown", "duration"]
dir = Path(__file__).parent
typespecs = [(dir / "namespace1.tsp", tests.idl.namespace1.expected_namespace)]


@pytest.mark.parametrize("type_symb,expected", type_symbols)
def test_type_t(type_symb: str, expected: type):
    assert (expected, "") == type_symbol(type_symb)


@pytest.mark.parametrize("path,expected", typespecs)
def test_namespace(path: Path, expected: Namespace):
    namespace = Namespace.from_spec_file(path)
    assert namespace.stubs() == expected.stubs()

    # Write stubs to file.
    # This file will be subject to linting.
    # Also a change in this file can indicate an unexpectedly breaking change.
    namespace.write_stubs(Path(__file__).parent / "stubs_output.py")


@pytest.mark.parametrize("type_symb", type_symbols_for_failure)
def test_parser_fail(type_symb: str):
    with pytest.raises(TierkreisError):
        type_symbol(type_symb)



from tierkreis.controller.data.core import EmptyModel
from tierkreis.controller.data.models import TKR, portmapping
from tierkreis.controller.data.types import Struct

__all__ = ["EmptyModel", "TKR", "portmapping", "Struct"]



from uuid import UUID
import pytest
from tests.controller.sample_graphdata import simple_eval, simple_map
from tierkreis.controller.data.core import PortID
from tierkreis.controller.data.graph import (
    Const,
    Eval,
    Func,
    GraphData,
    Input,
    graph_node_from_loc,
)
from tierkreis.controller.data.location import Loc
from tierkreis.controller.storage.graphdata import GraphDataStorage
from tierkreis.exceptions import TierkreisError


@pytest.mark.parametrize(
    ["node_location_str", "graph", "target"],
    [
        ("-.N0", simple_eval(), Const(0, outputs=set(["value"]))),
        ("-.N4.M0", simple_map(), Eval((-1, "body"), {})),
        ("-.N4.M0.N-1", simple_map(), Eval((-1, "body"), {})),
    ],
)
def test_read_nodedef(node_location_str: str, graph: GraphData, target: str) -> None:
    loc = Loc(node_location_str)
    storage = GraphDataStorage(UUID(int=0), graph)
    node_def = storage.read_node_def(loc)
    assert node_def == target


@pytest.mark.parametrize(
    ["node_location_str", "graph", "port", "target"],
    [
        ("-.N0", simple_eval(), "value", b"null"),
        ("-.N4.M0", simple_map(), "0", b"null"),
    ],
)
def test_read_output(
    node_location_str: str, graph: GraphData, port: PortID, target: str
) -> None:
    loc = Loc(node_location_str)
    storage = GraphDataStorage(UUID(int=0), graph)
    val = storage.read_output(loc, port)
    assert val == target


def test_raises() -> None:
    loc = Loc("-.N0")
    storage = GraphDataStorage(UUID(int=0), simple_eval())
    with pytest.raises(TierkreisError):
        storage.read_output(loc, "does_not_exist")


@pytest.mark.parametrize(
    ["node_location_str", "graph", "target"],
    [
        ("-.N0", simple_eval(), ["value"]),
        ("-.N4.M0", simple_map(), ["0"]),
    ],
)
def test_read_output_ports(
    node_location_str: str, graph: GraphData, target: str
) -> None:
    loc = Loc(node_location_str)
    storage = GraphDataStorage(UUID(int=0), graph)
    outputs = storage.read_output_ports(loc)
    assert outputs == target


@pytest.mark.parametrize(
    ["node_location_str", "graph", "target"],
    [
        ("-.N0", simple_eval(), Const(0, outputs=set(["value"]))),
        ("-.N3.N1", simple_eval(), Input("intercept", outputs=set(["intercept"]))),
        (
            "-.N3.N3",
            simple_eval(),
            Func(
                "builtins.itimes",
                inputs={"a": (0, "doubler_input"), "b": (2, "value")},
                outputs=set(["value"]),
            ),
        ),
        ("-.N-1", simple_eval(), Eval((-1, "body"), {})),
        ("-.N3.N-1", simple_eval(), Eval((-1, "body"), {})),
        (
            "-.N4.M0",
            simple_map(),
            Eval(
                (-1, "body"),
                inputs={"doubler_input": (2, "*"), "intercept": (0, "value")},
                outputs=set(["*"]),
            ),
        ),
        ("-.N4.M0.N-1", simple_map(), Eval((-1, "body"), {})),
        ("-.N4.M0.N1", simple_map(), Input("intercept", outputs=set(["intercept"]))),
    ],
)
def test_graph_node_from_loc(
    node_location_str: str, graph: GraphData, target: str
) -> None:
    loc = Loc(node_location_str)
    node_def, _ = graph_node_from_loc(loc, graph)
    assert node_def == target



from tierkreis.controller.data.graph import GraphData


def ternary_add() -> GraphData:
    g = GraphData()
    a = g.input("a")
    b = g.input("b")
    c = g.input("c")

    step_one = g.func("builtins.iadd", {"a": a, "b": b})("value")
    step_two = g.func("builtins.iadd", {"a": step_one, "b": c})("value")

    g.output({"ternary_add_output": step_two})
    return g


def double_partial() -> GraphData:
    g = GraphData()

    one = g.const(1)
    two = g.const(2)
    three = g.const(3)
    termary_add_const = g.const(ternary_add())

    add_add_1 = g.eval(termary_add_const, {"a": one})("body")
    add_3 = g.eval(add_add_1, {"b": two})("body")
    six = g.eval(add_3, {"c": three})("ternary_add_output")

    g.output({"double_partial_output": six})
    return g


def partial_intersection() -> GraphData:
    g = GraphData()
    one = g.const(1)
    two = g.const(2)
    three = g.const(3)
    termary_add_const = g.const(ternary_add())

    add_add_1 = g.eval(termary_add_const, {"a": one})("body")
    add_3 = g.eval(add_add_1, {"a": one, "b": two})("body")
    six = g.eval(add_3, {"c": three})("ternary_add_output")

    g.output({"double_partial_output": six})
    return g



import json
import pytest
from pathlib import Path
from uuid import UUID

from tierkreis import Labels
from tierkreis.controller.data.graph import (
    Const,
    EagerIfElse,
    Func,
    Input,
    Output,
)
from tests.controller.sample_graphdata import (
    simple_eagerifelse,
    simple_ifelse,
)

from tierkreis.controller import run_graph
from tierkreis.controller.data.location import Loc
from tierkreis.controller.data.types import PType
from tierkreis.controller.executor.shell_executor import ShellExecutor
from tierkreis.controller.executor.uv_executor import UvExecutor
from tierkreis.controller.storage.filestorage import ControllerFileStorage
from tierkreis.controller.data.graph import GraphData


def eagerifelse_long_running() -> GraphData:
    g = GraphData()
    pred = g.add(Input("pred"))("pred")
    pred_long = g.add(Func("controller.sleep_and_return", {"output": pred}))(
        Labels.VALUE
    )
    one = g.add(Const(1))(Labels.VALUE)
    one_long = g.add(Func("controller.sleep_and_return", {"output": one}))(Labels.VALUE)
    two = g.add(Const(2))(Labels.VALUE)
    out = g.add(EagerIfElse(pred_long, one_long, two))(Labels.VALUE)
    g.add(Output({"simple_eagerifelse_output": out}))
    return g


params = [({"pred": True}, 1), ({"pred": False}, 2)]


@pytest.mark.slow
@pytest.mark.parametrize("input, output", params)
def test_eagerifelse_long_running(input: dict[str, PType], output: int) -> None:
    g = eagerifelse_long_running()
    storage = ControllerFileStorage(UUID(int=150), name="eagerifelse_long_running")

    registry_path = Path(__file__).parent.parent
    executor = UvExecutor(registry_path=registry_path, logs_path=storage.logs_path)

    storage.clean_graph_files()
    run_graph(storage, executor, g, input, n_iterations=20000)
    actual_output = json.loads(storage.read_output(Loc(), "simple_eagerifelse_output"))
    assert actual_output == output


def test_eagerifelse_nodes() -> None:
    g = simple_eagerifelse()
    storage = ControllerFileStorage(UUID(int=150), name="simple_if_else")
    executor = ShellExecutor(Path("./python/examples/launchers"), storage.workflow_dir)
    storage.clean_graph_files()
    run_graph(storage, executor, g, {"pred": b"true"})
    assert storage.is_node_finished(Loc("-.N3"))
    assert storage.is_node_finished(Loc("-.N4"))


def test_ifelse_nodes():
    g = simple_ifelse()
    storage = ControllerFileStorage(UUID(int=151), name="simple_if_else")
    executor = ShellExecutor(Path("./python/examples/launchers"), storage.workflow_dir)
    storage.clean_graph_files()
    run_graph(storage, executor, g, {"pred": b"true"})
    assert storage.is_node_finished(Loc("-.N1"))
    assert not storage.is_node_finished(Loc("-.N2"))



from logging import getLogger
from pathlib import Path
from typing import Any, Literal, Optional

from pydantic import BaseModel, GetCoreSchemaHandler
from pydantic_core import CoreSchema, core_schema
from tierkreis.controller.data.core import PortID
from typing_extensions import assert_never

from tierkreis.controller.data.core import NodeIndex
from tierkreis.exceptions import TierkreisError

logger = getLogger(__name__)


class WorkerCallArgs(BaseModel):
    function_name: str
    inputs: dict[str, Path]
    outputs: dict[str, Path]
    output_dir: Path
    done_path: Path
    error_path: Path
    logs_path: Optional[Path]


NodeStep = Literal["-"] | tuple[Literal["N", "L", "M"], NodeIndex]


class Loc(str):
    def __new__(cls, k: str = "-") -> "Loc":
        return super(Loc, cls).__new__(cls, k)

    def N(self, idx: int) -> "Loc":
        return Loc(str(self) + f".N{idx}")

    def L(self, idx: int) -> "Loc":
        return Loc(str(self) + f".L{idx}")

    def M(self, idx: int) -> "Loc":
        return Loc(str(self) + f".M{idx}")

    @staticmethod
    def from_steps(steps: list[NodeStep]) -> "Loc":
        loc = ""
        for step in steps.copy():
            match step:
                case "-":
                    loc += "-"
                case (node_type, idx):
                    loc += f".{node_type}{idx}"
        return Loc(loc)

    def parent(self) -> "Loc | None":
        steps = self.steps()
        if not steps:
            return None

        last_step = steps.pop()
        match last_step:
            case "-":
                return Loc.from_steps([])
            case ("L", 0):
                return Loc.from_steps(steps)
            case ("L", idx):
                return Loc.from_steps(steps).L(idx - 1)
            case ("N", idx) | ("M", idx):
                return Loc.from_steps(steps)
            case _:
                assert_never(last_step)

    def steps(self) -> list[NodeStep]:
        if self == "":
            return []

        steps: list[NodeStep] = []
        for step_str in self.split("."):
            match step_str[0], step_str[1:]:
                case ("-", _):
                    steps.append("-")
                case ("N", idx_str):
                    steps.append(("N", int(idx_str)))
                case ("L", idx_str):
                    steps.append(("L", int(idx_str)))
                case ("M", idx_str):
                    steps.append(("M", int(idx_str)))
                case _:
                    raise TierkreisError(f"Invalid Loc: {self}")

        return steps

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: GetCoreSchemaHandler
    ) -> CoreSchema:
        return core_schema.no_info_after_validator_function(cls, handler(str))

    def pop_first(self) -> tuple[NodeStep, "Loc"]:
        if self == "-":
            return "-", Loc("")
        steps = self.steps()
        if len(steps) < 2:
            raise TierkreisError("Malformed Loc")
        first = steps.pop(1)
        if first == "-":
            raise TierkreisError("Malformed Loc")
        return first, Loc.from_steps(steps)

    def pop_last(self) -> tuple[NodeStep, "Loc"]:
        if self == "-":
            return "-", Loc("")
        steps = self.steps()
        if len(steps) < 2:
            raise TierkreisError("Malformed Loc")
        last = steps.pop(-1)
        if last == "-":
            raise TierkreisError("Malformed Loc")
        return last, Loc.from_steps(steps)


def get_last_index(loc: Loc) -> int:
    step = loc.steps()[-1]
    if isinstance(step, str):
        return 0
    return step[1]


OutputLoc = tuple[Loc, PortID]



"""Code generated from TestNamespace namespace. Please do not edit."""

from typing import NamedTuple, Protocol
from tierkreis.controller.data.models import TKR
from tierkreis.controller.data.types import PType, Struct


class A(NamedTuple):
    age: TKR[int]  # noqa: F821 # fmt: skip
    name: TKR[dict[str, str]]  # noqa: F821 # fmt: skip


class B(Struct, Protocol):
    age: int  # noqa: F821 # fmt: skip
    name: dict[str, str]  # noqa: F821 # fmt: skip


class C[T: PType](Struct, Protocol):
    a: list[int]  # noqa: F821 # fmt: skip
    b: B  # noqa: F821 # fmt: skip
    t: T  # noqa: F821 # fmt: skip


class foo(NamedTuple):
    a: TKR[int]  # noqa: F821 # fmt: skip
    b: TKR[str]  # noqa: F821 # fmt: skip

    @staticmethod
    def out() -> type[A]:  # noqa: F821 # fmt: skip
        return A  # noqa: F821 # fmt: skip

    @property
    def namespace(self) -> str:
        return "TestNamespace"


class bar(NamedTuple):
    @staticmethod
    def out() -> type[TKR[B]]:  # noqa: F821 # fmt: skip
        return TKR[B]  # noqa: F821 # fmt: skip

    @property
    def namespace(self) -> str:
        return "TestNamespace"


class z[T: PType](NamedTuple):
    c: TKR[C[T]]  # noqa: F821 # fmt: skip

    @staticmethod
    def out() -> type[TKR[C[T]]]:  # noqa: F821 # fmt: skip
        return TKR[C[T]]  # noqa: F821 # fmt: skip

    @property
    def namespace(self) -> str:
        return "TestNamespace"



import logging
import importlib.util
from pathlib import Path

from tierkreis.controller.data.location import Loc
from tierkreis.controller.storage.in_memory import ControllerInMemoryStorage
from tierkreis.worker.storage.in_memory import InMemoryWorkerStorage
from tierkreis.exceptions import TierkreisError


logger = logging.getLogger(__name__)


class InMemoryExecutor:
    """Executes workers in the same process as the controller.

    Implements: :py:class:`tierkreis.controller.executor.protocol.ControllerExecutor`
    """

    def __init__(self, registry_path: Path, storage: ControllerInMemoryStorage) -> None:
        self.registry_path = registry_path
        self.storage = storage

    def run(
        self,
        launcher_name: str,
        worker_call_args_path: Path,
    ) -> None:
        logging.basicConfig(
            format="%(asctime)s: %(message)s",
            datefmt="%Y-%m-%dT%H:%M:%S%z",
            filemode="a",
            level=logging.INFO,
        )
        logger.info("START %s %s", launcher_name, worker_call_args_path)
        node_location = Loc(str(worker_call_args_path))
        call_args = self.storage.read_worker_call_args(node_location)

        spec = importlib.util.spec_from_file_location(
            "in_memory", self.registry_path / launcher_name / "main.py"
        )
        if spec is None or spec.loader is None:
            raise TierkreisError(
                f"Couldn't load main.py in {self.registry_path / launcher_name}"
            )
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        worker_storage = InMemoryWorkerStorage(self.storage)
        module.worker.storage = worker_storage
        module.worker.functions[call_args.function_name](call_args)
        self.storage.mark_node_finished(node_location)



# ruff: noqa: F821
from typing import NamedTuple
from tierkreis.builder import GraphBuilder
from tierkreis.builtins.stubs import tkr_sleep
from tierkreis.controller.data.models import TKR, OpaqueType
from tierkreis.nexus_worker import (
    upload_circuit,
    start_execute_job,
    is_running,
    get_results,
)

type Circuit = OpaqueType["pytket._tket.circuit.Circuit"]
type BackendResult = OpaqueType["pytket.backends.backendresult.BackendResult"]
type ExecuteJobRef = OpaqueType["qnexus.models.references.ExecuteJobRef"]
type ExecutionProgram = OpaqueType["qnexus.models.references.ExecuteJobRef"]


class UploadCircuitInputs(NamedTuple):
    project_name: TKR[str]
    circuit: TKR[Circuit]


class JobInputs(NamedTuple):
    project_name: TKR[str]
    job_name: TKR[str]
    circuits: TKR[list[Circuit]]
    n_shots: TKR[list[int]]
    backend_config: TKR[OpaqueType["qnexus.BackendConfig"]]


class LoopOutputs(NamedTuple):
    results: TKR[list[BackendResult]]
    should_continue: TKR[bool]


def upload_circuit_graph():
    g = GraphBuilder(UploadCircuitInputs, TKR[ExecutionProgram])
    programme = g.task(upload_circuit(g.inputs.project_name, g.inputs.circuit))
    g.outputs(programme)  # type: ignore
    return g


def polling_loop_body(polling_interval: float):
    g = GraphBuilder(TKR[ExecuteJobRef], LoopOutputs)
    pred = g.task(is_running(g.inputs))

    wait = g.ifelse(pred, g.task(tkr_sleep(g.const(polling_interval))), g.const(False))
    results = g.ifelse(pred, g.const([]), g.task(get_results(g.inputs)))

    g.outputs(LoopOutputs(results=results, should_continue=wait))
    return g


def nexus_submit_and_poll(polling_interval: float = 30.0):
    g = GraphBuilder(JobInputs, TKR[list[BackendResult]])
    upload_inputs = g.map(
        lambda x: UploadCircuitInputs(g.inputs.project_name, x), g.inputs.circuits
    )
    programmes = g.map(upload_circuit_graph(), upload_inputs)

    ref = g.task(
        start_execute_job(
            g.inputs.project_name,
            g.inputs.job_name,
            programmes,  # type: ignore
            g.inputs.n_shots,
            g.inputs.backend_config,  # type: ignore
        )
    )

    res = g.loop(polling_loop_body(polling_interval), ref)
    g.outputs(res.results)
    return g



from dataclasses import dataclass
from logging import getLogger
import logging
from pathlib import Path
import subprocess
import sys

from tierkreis.controller.data.core import PortID
from tierkreis.controller.data.types import bytes_from_ptype, ptype_from_bytes
from tierkreis.controller.executor.in_memory_executor import InMemoryExecutor
from tierkreis.controller.storage.adjacency import outputs_iter
from typing_extensions import assert_never

from tierkreis.consts import PACKAGE_PATH
from tierkreis.controller.data.graph import Eval, GraphData, NodeDef
from tierkreis.controller.data.location import Loc, OutputLoc
from tierkreis.controller.executor.protocol import ControllerExecutor
from tierkreis.controller.storage.protocol import ControllerStorage
from tierkreis.controller.storage.in_memory import ControllerInMemoryStorage
from tierkreis.labels import Labels
from tierkreis.exceptions import TierkreisError

logger = logging.getLogger(__name__)


@dataclass
class NodeRunData:
    node_location: Loc
    node: NodeDef
    output_list: list[PortID]


def start_nodes(
    storage: ControllerStorage,
    executor: ControllerExecutor,
    node_run_data: list[NodeRunData],
) -> None:
    started_locs: set[Loc] = set()
    for node_run_datum in node_run_data:
        if node_run_datum.node_location in started_locs:
            continue
        start(storage, executor, node_run_datum)
        started_locs.add(node_run_datum.node_location)


def run_builtin(def_path: Path, logs_path: Path) -> None:
    logger = getLogger("builtins")
    if not logger.hasHandlers():
        formatter = logging.Formatter(
            fmt="%(asctime)s: %(message)s",
            datefmt="%Y-%m-%dT%H:%M:%S%z",
        )
        handler = logging.FileHandler(logs_path, mode="a")
        handler.setFormatter(formatter)
        logger.setLevel(logging.INFO)

        logger.addHandler(handler)

    logger.info("START builtin %s", def_path)
    with open(logs_path, "a") as fh:
        subprocess.Popen(
            [sys.executable, "main.py", def_path],
            start_new_session=True,
            cwd=PACKAGE_PATH / "tierkreis" / "builtins",
            stderr=fh,
            stdout=fh,
        )


def start(
    storage: ControllerStorage, executor: ControllerExecutor, node_run_data: NodeRunData
) -> None:
    node_location = node_run_data.node_location
    node = node_run_data.node
    output_list = node_run_data.output_list

    storage.write_node_def(node_location, node)

    parent = node_location.parent()
    if parent is None:
        raise TierkreisError(f"{node.type} node must have parent Loc.")

    ins = {k: (parent.N(idx), p) for k, (idx, p) in node.inputs.items()}

    logger.debug(f"start {node_location} {node} {ins} {output_list}")
    if node.type == "function":
        name = node.function_name
        launcher_name = ".".join(name.split(".")[:-1])
        name = name.split(".")[-1]
        call_args_path = storage.write_worker_call_args(
            node_location, name, ins, output_list
        )
        logger.debug(f"Executing {(str(node_location), name, ins, output_list)}")

        if isinstance(storage, ControllerInMemoryStorage) and isinstance(
            executor, InMemoryExecutor
        ):
            executor.run(launcher_name, call_args_path)

        elif launcher_name == "builtins":
            run_builtin(call_args_path, storage.logs_path)
        else:
            executor.run(launcher_name, call_args_path)

    elif node.type == "input":
        input_loc = parent.N(-1)
        storage.link_outputs(node_location, node.name, input_loc, node.name)
        storage.mark_node_finished(node_location)

    elif node.type == "output":
        storage.mark_node_finished(node_location)

        pipe_inputs_to_output_location(storage, parent, ins)
        storage.mark_node_finished(parent)

    elif node.type == "const":
        bs = bytes_from_ptype(node.value)
        storage.write_output(node_location, Labels.VALUE, bs)
        storage.mark_node_finished(node_location)

    elif node.type == "eval":
        message = storage.read_output(parent.N(node.graph[0]), node.graph[1])
        g = ptype_from_bytes(message, GraphData)
        ins["body"] = (parent.N(node.graph[0]), node.graph[1])
        ins.update(g.fixed_inputs)

        pipe_inputs_to_output_location(storage, node_location.N(-1), ins)

    elif node.type == "loop":
        ins["body"] = (parent.N(node.body[0]), node.body[1])
        pipe_inputs_to_output_location(storage, node_location.N(-1), ins)
        start(
            storage,
            executor,
            NodeRunData(
                node_location.L(0),
                Eval((-1, "body"), {k: (-1, k) for k, _ in ins.items()}, node.outputs),
                output_list,
            ),
        )

    elif node.type == "map":
        first_ref = next(x for x in ins.values() if x[1] == "*")
        map_eles = outputs_iter(storage, first_ref[0])
        for idx, p in map_eles:
            eval_inputs: dict[PortID, tuple[Loc, PortID]] = {}
            eval_inputs["body"] = (parent.N(node.body[0]), node.body[1])
            for k, (i, port) in ins.items():
                if port == "*":
                    eval_inputs[k] = (i, p)
                else:
                    eval_inputs[k] = (i, port)
            pipe_inputs_to_output_location(
                storage, node_location.M(idx).N(-1), eval_inputs
            )
            # Necessary in the node visualization
            storage.write_node_def(
                node_location.M(idx), Eval((-1, "body"), node.inputs, node.outputs)
            )

    elif node.type == "ifelse":
        pass

    elif node.type == "eifelse":
        pass
    else:
        assert_never(node)


def pipe_inputs_to_output_location(
    storage: ControllerStorage,
    output_loc: Loc,
    inputs: dict[PortID, OutputLoc],
) -> None:
    for new_port, (old_loc, old_port) in inputs.items():
        storage.link_outputs(output_loc, new_port, old_loc, old_port)



from dataclasses import dataclass
from inspect import isclass
from itertools import chain
from typing import (
    Literal,
    Protocol,
    Union,
    cast,
    dataclass_transform,
    get_args,
    get_origin,
    overload,
    runtime_checkable,
)
from typing_extensions import TypeIs
from tierkreis.controller.data.core import (
    NodeIndex,
    PortID,
    RestrictedNamedTuple,
    ValueRef,
)
from tierkreis.controller.data.types import PType, generics_in_ptype

TKR_PORTMAPPING_FLAG = "__tkr_portmapping__"


@runtime_checkable
class PNamedModel(RestrictedNamedTuple[PType], Protocol):
    """A struct whose members are restricted to being PTypes.

    E.g. used to specify multiple outputs in Python worker code.
    """


@dataclass_transform()
def portmapping[T: PNamedModel](cls: type[T]) -> type[T]:
    setattr(cls, TKR_PORTMAPPING_FLAG, True)
    return cls


PModel = PNamedModel | PType
OpaqueType = Literal


@dataclass
class TKR[T: PModel]:
    node_index: NodeIndex
    port_id: PortID

    def value_ref(self) -> ValueRef:
        return (self.node_index, self.port_id)


@runtime_checkable
class TNamedModel(RestrictedNamedTuple[TKR[PType] | None], Protocol):
    """A struct whose members are restricted to being references to PTypes.

    E.g. in graph builder code these are outputs of tasks."""


TModel = TNamedModel | TKR


@overload
def is_portmapping(o: PModel) -> TypeIs[PNamedModel]: ...
@overload
def is_portmapping(o: type[PModel]) -> TypeIs[type[PNamedModel]]: ...
@overload
def is_portmapping(o: TModel) -> TypeIs[TNamedModel]: ...
@overload
def is_portmapping(o: type[TModel]) -> TypeIs[type[TNamedModel]]: ...
def is_portmapping(
    o,
) -> (
    TypeIs[type[PNamedModel]]
    | TypeIs[PNamedModel]
    | TypeIs[TNamedModel]
    | TypeIs[type[TNamedModel]]
):
    origin = get_origin(o)
    if origin is not None:
        return is_portmapping(origin)
    return hasattr(o, TKR_PORTMAPPING_FLAG)


def is_tnamedmodel(o) -> TypeIs[type[TNamedModel]]:
    origin = get_origin(o)
    if origin is not None:
        return is_tnamedmodel(origin)
    return isclass(o) and issubclass(o, TNamedModel)


def dict_from_pmodel(pmodel: PModel) -> dict[PortID, PType]:
    if is_portmapping(pmodel):
        return pmodel._asdict()

    return {"value": pmodel}


def dict_from_tmodel(tmodel: TModel) -> dict[PortID, ValueRef]:
    if isinstance(tmodel, TNamedModel):
        return {k: (v.node_index, v.port_id) for k, v in tmodel._asdict().items() if v}

    return {"value": (tmodel.node_index, tmodel.port_id)}


def model_fields(model: type[PModel] | type[TModel]) -> list[str]:
    if is_portmapping(model):
        return getattr(model, "_fields")

    if is_tnamedmodel(model):
        return getattr(model, "_fields")

    return ["value"]


def init_tmodel[T: TModel](tmodel: type[T], refs: list[ValueRef]) -> T:
    if is_tnamedmodel(tmodel):
        o = get_origin(tmodel)
        model = tmodel if not is_tnamedmodel(o) else o
        args: list[TKR] = []
        for ref in refs:
            key = ref[1].replace("-*", "")
            param = model.__annotations__[key]
            if get_origin(param) == Union:
                param = next(x for x in get_args(param) if x)
            args.append(param(ref[0], ref[1]))
        return cast(T, model(*args))
    return tmodel(refs[0][0], refs[0][1])


def generics_in_pmodel(pmodel: type[PModel]) -> set[str]:
    if is_portmapping(pmodel):
        origin = get_origin(pmodel)
        if origin is not None:
            return generics_in_pmodel(origin)

        x = [generics_in_ptype(pmodel.__annotations__[t]) for t in model_fields(pmodel)]
        return set(chain(*x))

    return generics_in_ptype(pmodel)
